import enum
from abc import ABC, abstractmethod
from typing import List

from qrogue.game.logic.collectibles import Collectible


class ControllableType(enum.Enum):
    """
    Specifies the different types of Controllable for automatic processing (e.g. for color rules).
    """

    Luke = "L"
    Test = "T"
    Base = "Q"

    @staticmethod
    def values() -> List["ControllableType"]:
        """

        :return: list of all ControllableTypes
        """
        return [ControllableType.Luke, ControllableType.Test, ControllableType.Base]

    def __init__(self, abbreviation: str):
        """
        Initializes the ControllableType with a (unique) abbreviation.

        :param abbreviation: how the Controllable is displayed
        """
        self.__abbreviation = abbreviation

    @property
    def name(self) -> str:
        return self.__abbreviation


class Controllable(ABC):
    """
    A Controllable is the interface between the game's player and the game.
    """

    def __init__(self, name: str):  # todo use ControllableType instead
        self.__name = name

    @property
    def name(self) -> str:
        """

        :return: name of the Controllable
        """
        return self.__name

    @abstractmethod
    def game_over_check(self) -> bool:
        """
        Checks whether the Controllable is game over or not.

        :return: True if the Controllable is game over, False otherwise
        """
        pass

    @abstractmethod
    def key_count(self) -> int:
        """

        :return: how many keys the Controllable can use
        """
        pass

    @abstractmethod
    def use_key(self) -> bool:
        """
        Tries to use one of the Controllable's keys.

        :return: True if a key was used successfully, False otherwise
        """
        pass

    @abstractmethod
    def get_img(self):
        """
        Provides an image for rendering. Return type depends on the renderer we use.

        :return: an image to render the Controllable
        """
        pass

    @abstractmethod
    def description(self) -> str:
        """
        Provides a description to display some prose information about the Controllable. E.g. can be used to show the
        game's player details about the Controllable.

        :return: a short description of the Controllable
        """
        pass

    @abstractmethod
    def give_collectible(self, collectible: Collectible, force: bool = False) -> bool:
        """
        Gives the provided Collectible to the Controllable (e.g. adding it to its inventory).

        :param collectible: a Collectible to give the Controllable
        :param force: whether we force collectible to be added, defaults to False
        :return: True if the collectible was given successfully, False otherwise
        """
        pass

    def __str__(self) -> str:
        return f"{self.name} [\"{self.description()}\"]"



from typing import List, Callable, Optional

from qrogue.game.logic.actors import Robot
from qrogue.game.world.navigation import Coordinate
from .map import Map, MapType, Room, MapMetaData


class LessonMap(Map):
    def __init__(self, meta_data: MapMetaData, file_name: str, seed: int, rooms: List[List[Optional[Room]]],
                 robot: Robot, spawn_room: Coordinate, check_achievement_callback: Callable[[str], bool],
                 trigger_event_callback: Callable[[str], None]):
        super().__init__(meta_data, file_name, seed, rooms, robot, spawn_room, check_achievement_callback,
                         trigger_event_callback)

    def get_type(self) -> MapType:
        return MapType.Level



from abc import ABC
from typing import Optional, Callable

from qrogue.game.logic.collectibles import Collectible, CollectibleType
from qrogue.util import Logger


class Pickup(Collectible, ABC):
    def __init__(self, amount: int, type_: CollectibleType = CollectibleType.Pickup):
        super(Pickup, self).__init__(type_)
        if amount <= 0:
            amount = 1
        self._amount = amount

    @property
    def amount(self):
        return self._amount

    def __str__(self):
        return self.to_string()


class Score(Pickup):
    def __init__(self, amount: int = 100):
        if amount < 0:
            Logger.instance().warn(f"Negative amount (={amount}) defined! Will use 0 instead.")
            amount = 0
        super().__init__(amount, type_=CollectibleType.Score)

    def name(self) -> str:
        return "Score"

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        return "Score describes how well you performed in a level."

    def to_string(self) -> str:
        return f"Score #{self.amount}"


class Key(Pickup):
    def __init__(self, amount: int = 1):
        super().__init__(amount, type_=CollectibleType.Key)

    def name(self) -> str:
        return "Key"

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        return "Keys are useful for opening locked doors."

    def to_string(self):
        if self.amount > 1:
            return f"{self.amount} keys"
        return f"{self.amount} key"


class Energy(Pickup):
    def __init__(self, amount: int = 10):
        super().__init__(amount, type_=CollectibleType.Energy)

    def name(self) -> str:
        return "Energy"

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        return "Gives back some energy to the Robot so it can stay longer on a mission."

    def to_string(self) -> str:
        return f"Energy ({self.amount})"


class Qubit(Collectible):
    __MIN_NUM = 1
    __MAX_NUM = 3

    def __init__(self, num: int = 1):
        if self.__MIN_NUM <= num <= self.__MAX_NUM:
            super().__init__(CollectibleType.Qubit)
            self.__num = num
        else:
            raise ValueError(f"Error trying to add to many qubits: {num}! Only between {self.__MIN_NUM} and "
                             f"{self.__MAX_NUM} allowed at once.")

    @property
    def additional_qubits(self) -> int:
        return self.__num

    def name(self) -> str:
        return "Qubit"

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        return "Qubits are the essence of any circuit you create."

    def to_string(self) -> str:
        return f"Qubit ({self.additional_qubits})"



import time
from threading import Thread
from typing import Callable, Optional, Dict, List, Tuple, Union

from qrogue.game.logic import Message
from qrogue.game.logic.actors.controllables.robot import RoboProperties
from qrogue.game.logic.collectibles import Instruction
from qrogue.game.world.dungeon_generator import ExpeditionGenerator, QrogueLevelGenerator
from qrogue.game.world.dungeon_generator.wave_function_collapse import WFCManager
from qrogue.game.world.map import Map, MapType, ExpeditionMap, CallbackPack
from qrogue.game.world.navigation import Coordinate
from qrogue.graphics.popups import Popup
from qrogue.management.save_data import NewSaveData
from qrogue.util import CommonQuestions, RandomManager, LevelInfo, Config, MapConfig, ErrorConfig, \
    ExpeditionConfig, StvDifficulty
from qrogue.util.achievements import Achievement
from qrogue.util.util_functions import cur_datetime, time_diff


class MapManager:
    @staticmethod
    def parse_expedition_parameters(map_name: str, map_seed: Optional[int], expedition_progress: int) \
            -> Tuple[str, int, Optional[int]]:
        # Example: expedition0d7p11 -> Difficulty = 0, map seed = 7, puzzle seed = 11
        # first: extract difficulty and seed information from map_name if available
        if len(map_name) > len(MapConfig.expedition_map_prefix()):
            # map_name contains additional information
            info = map_name[len(MapConfig.expedition_map_prefix()):]

            def parse_seed_info(_seed_info: str) -> Tuple[Optional[int], int]:
                # seed info is either solely for map_seed or for both type of seeds
                if MapConfig.puzzle_seed_separator() in _seed_info:
                    # info_seed contains information about both seeds
                    _info_puzzle, _info_map = _seed_info.split(MapConfig.puzzle_seed_separator())
                    return int(_info_puzzle), int(_info_map)
                else:
                    # _seed_info only contains information about map_seed
                    return None, int(_seed_info)

            if MapConfig.diff_code_separator() in info:
                diff_code, info_seed = info.split(MapConfig.diff_code_separator())
                # info contains both difficulty and seed information
                if map_seed is None or MapConfig.puzzle_seed_separator() in info_seed:
                    puzzle_seed, map_seed = parse_seed_info(info_seed)
                else:
                    # seed info is solely for puzzle_seed
                    puzzle_seed = int(info_seed)
            elif map_seed is None:
                # there is no difficulty code, just seed information (map_seed is implicitly None due to 1st if)
                expedition_progress = int(expedition_progress)
                diff_code = str(LevelInfo.get_expedition_difficulty(expedition_progress))
                puzzle_seed, map_seed = parse_seed_info(info)
                Config.check_reachability("load_map(): expedition without seed")
            else:
                # a map_seed was provided and there is no diff_code separator -> info is solely for diff_code
                diff_code = info
                puzzle_seed = None
        else:
            # no additional information given, hence, we use default values
            diff_code = str(LevelInfo.get_expedition_difficulty(expedition_progress))
            puzzle_seed = None

        assert map_seed is not None, "MapSeed is None after parsing expedition arguments!"
        return diff_code, map_seed, puzzle_seed

    def __init__(self, wfc_manager: WFCManager, save_data: NewSaveData, seed: int, start_level: Callable[[Map], None],
                 start_level_transition_callback: Callable[[str, str, Callable[[], None]], None],
                 exit_map_callback: Callable[[], None], callback_pack: CallbackPack,
                 queue_size: int = ExpeditionConfig.DEFAULT_QUEUE_SIZE):
        self.__save_data = save_data
        # no longer used code!
        self.__exit_map = exit_map_callback
        self.__cbp = callback_pack
        self.__queue_size = queue_size

        self.__rm = RandomManager.create_new(seed)
        self.__start_level = start_level
        self.__start_level_transition = start_level_transition_callback
        self.__expedition_generator = ExpeditionGenerator(wfc_manager, self.__save_data.check_achievement,
                                                          self.trigger_event, self.__load_back,
                                                          save_data.get_gates, callback_pack)  # todo: pass ExpeditionGenerator as argument?
        self.__expedition_queue: List[ExpeditionMap] = []
        self.__cur_map: Optional[Map] = None

        self.__level_timer = cur_datetime()
        self.__temp_level_event_storage: Dict[str, Tuple[int, int]] = {}  # event name -> score, done_score

    @property
    def is_in_level(self) -> bool:
        return self.__cur_map is not None and self.__cur_map.get_type() is MapType.Level

    @property
    def is_in_expedition(self) -> bool:
        return self.__cur_map is not None and self.__cur_map.get_type() is MapType.Expedition

    @property
    def show_individual_qubits(self) -> bool:
        return self.__cur_map.show_individual_qubits

    def fill_expedition_queue(self, callback: Optional[Callable[[], None]] = None, no_thread: bool = False):
        if len(self.__expedition_queue) >= self.__queue_size:
            return

        def fill():
            # todo: how to handle difficulty?
            difficulty = StvDifficulty.from_difficulty_code("1")
            robo_props = RoboProperties.from_difficulty_level(difficulty.level)
            while len(self.__expedition_queue) < self.__queue_size:
                map_seed = self.__rm.get_seed("MapManager.fill()@map_seed")
                puzzle_seed = self.__rm.get_seed("MapManager.fill()@puzzle_seed")
                expedition, success = self.__expedition_generator.generate(map_seed,
                                                                           (robo_props, difficulty, puzzle_seed))
                if success:
                    self.__expedition_queue.append(expedition)

            if callback is not None:
                callback()

        if no_thread:
            fill()
        else:
            Thread(target=fill, args=(), daemon=True).start()

    def __load_level(self, map_name: str, spawn_room: Optional[Coordinate] = None, seed: Optional[int] = None,
                     gate_list: Optional[List[Instruction]] = None):
        # generate a random seed for all paths since the randomizer state should not depend on whether we passed a seed
        # manually (e.g., during debugging) or not
        rand_seed = self.__rm.get_seed("MapManager.__load_level()@seedForLevel")
        if seed is None: seed = rand_seed

        # todo maybe levels should be able to have arbitrary names except "w..." or "e..." or "back" or "next"?
        check_achievement = self.__save_data.check_achievement
        generator = QrogueLevelGenerator(check_achievement, self.trigger_event, self.load_map, Popup.npc_says,
                                         self.__cbp)
        try:
            level, success = generator.generate(seed, map_name)
            if success:
                self.__cur_map = level
                # if we have a valid gate_list, we try to set it and only show the error if it fails
                if gate_list is not None and not self.__cur_map.robot.set_available_instructions(gate_list):
                    Popup.error(f"Failed to set gates! Please make sure you don't have more than "
                                f"{self.__cur_map.robot.capacity} gates selected.")
                # else: available gates were determined by the level beforehand and don't change
                self.__start_level(self.__cur_map)
            else:
                Popup.error(f"Failed to generate level \"{map_name}\"!", add_report_note=True)
        except FileNotFoundError:
            Popup.error(ErrorConfig.invalid_map(map_name, f"Level-file for \"{map_name}\" was not found! "),
                        add_report_note=True)

    def __load_expedition(self, difficulty: Union[StvDifficulty, str], map_seed: Optional[int] = None,
                          puzzle_seed: Optional[int] = None, gate_list: Optional[List[Instruction]] = None) -> None:
        # generate random seeds for all paths since the randomizer state should not depend on whether we passed a seed
        # manually (e.g., during debugging) or not
        rand_map_seed = self.__rm.get_seed("MapManager.__load_expedition()@map_seed")
        rand_puzzle_seed = self.__rm.get_seed("MapManager.__load_expedition()@puzzle_seed")

        if isinstance(difficulty, str):
            difficulty = StvDifficulty.from_difficulty_code(difficulty)
        robo_props = RoboProperties.from_difficulty_level(difficulty.level, gate_list)

        if map_seed is None and self.__queue_size > 0:
            while len(self.__expedition_queue) <= 0:
                time.sleep(Config.loading_refresh_time())
            success = True
            expedition = self.__expedition_queue.pop(0)
            self.fill_expedition_queue()  # todo: how to handle difficulty? dictionary?
            Config.check_reachability("post queue filling")
        else:
            if map_seed is None: map_seed = rand_map_seed
            if puzzle_seed is None: puzzle_seed = rand_puzzle_seed
            expedition, success = self.__expedition_generator.generate(map_seed, (robo_props, difficulty,
                                                                                  puzzle_seed))

        if success:
            self.__cur_map = expedition
            self.__start_level(self.__cur_map)
        else:
            Popup.error(f"Failed to create an expedition for seed = {map_seed}. Please try again with a different "
                        f"seed or restart the game. Should the error keep occurring:", add_report_note=True)

    def load_map(self, map_name: str, spawn_room: Optional[Coordinate] = None, seed: Optional[int] = None,
                 gate_list: Optional[List[Instruction]] = None):
        if map_name == MapConfig.first_uncleared():
            next_map = LevelInfo.get_next(MapConfig.first_uncleared(), self.__save_data.check_level)
            if next_map is None:
                Popup.error(f"Failed to find the next map of \"{map_name}\". Please restart and try again.\n"
                            f"If this error still occurs but you're sure that the corresponding file is present:",
                            add_report_note=True)
            else:
                self.load_map(next_map, spawn_room, seed, gate_list)

        elif map_name.lower() == MapConfig.next_map_string():
            self.__load_next()

        elif map_name.lower() == MapConfig.back_map_string():
            self.__load_back()

        elif map_name.lower().startswith(MapConfig.lesson_map_prefix()):
            self.__load_level(map_name, spawn_room, seed, gate_list)

        elif map_name.lower().startswith(MapConfig.expedition_map_prefix()):
            expedition_progress = int(self.__save_data.get_progress(Achievement.CompletedExpedition)[0])
            diff_code, map_seed, puzzle_seed = MapManager.parse_expedition_parameters(map_name, seed,
                                                                                      expedition_progress)
            # load expedition
            self.__load_expedition(diff_code, map_seed, puzzle_seed, gate_list)

        else:
            Popup.error(ErrorConfig.invalid_map(map_name), add_report_note=True)

    def __load_next(self):
        # generate random seeds for all paths since the randomizer state should not depend on whether we passed a seed
        # manually (e.g., during debugging) or not
        rand_map_seed = self.__rm.get_seed("MapManager.load_next()@expedition>map_seed")
        rand_puzzle_seed = self.__rm.get_seed("MapManager.load_next()@expedition>puzzle_seed")

        next_map = LevelInfo.get_next(self.__cur_map.internal_name, self.__save_data.check_level)
        if next_map is None:
            error_text = ErrorConfig.invalid_map(self.__cur_map.name, f"Failed to load next map after "
                                                                      f"\"{self.__cur_map.name}\". ")
            Popup.error(error_text, add_report_note=True)

        elif next_map.startswith(MapConfig.expedition_map_prefix()):
            # create difficulty code based on expedition progress
            expedition_progress = int(self.__save_data.get_progress(Achievement.CompletedExpedition)[0])
            diff_code = str(LevelInfo.get_expedition_difficulty(expedition_progress))

            difficulty = StvDifficulty.from_difficulty_code(diff_code)
            self.__start_level_transition(self.__cur_map.name, ExpeditionMap.to_display_name(difficulty, rand_map_seed),
                                          lambda: self.__load_expedition(difficulty, rand_map_seed, rand_puzzle_seed))
        else:
            self.__start_level_transition(self.__cur_map.name, LevelInfo.convert_to_display_name(next_map),
                                          lambda: self.load_map(next_map, None, rand_map_seed))

    def __load_back(self):
        self.__exit_map()

    def __proceed(self, confirmed: int = 0):
        # not defined inside trigger_event() since it doesn't depend on any internal state of trigger_event() and,
        #  therefore, it's unnecessary to create __proceed again for every trigger_event()-call
        if confirmed == 0:
            self.__load_next()
        elif confirmed == 1:
            pass  # stay
        elif confirmed == 2:
            self.__load_back()

    def trigger_event(self, event_id: str):
        if event_id.lower() == MapConfig.done_event_id():
            # event_id = MapConfig.specific_done_event_id(self.__cur_map.internal_name)
            # self.__save_data.trigger_event(event_id)
            prev_level_data = self.__save_data.get_level_data(self.__cur_map.internal_name)
            duration, _ = time_diff(cur_datetime(), self.__level_timer)

            # store progress in save_data
            if self.is_in_level:
                new_level_data = self.__save_data.complete_level(
                    self.__cur_map.internal_name, duration, score=self.__cur_map.robot.score)
            elif self.is_in_expedition:
                assert isinstance(self.__cur_map, ExpeditionMap), \
                    f"CurMap has wrong type: \"{self.__cur_map.internal_name}\" is no ExpeditionMap!"
                new_level_data = self.__save_data.complete_expedition(
                    self.__cur_map.internal_name, duration, self.__cur_map.difficulty.level, self.__cur_map.main_gate,
                    score=self.__cur_map.robot.score)
            else:   # currently only levels and expedition can trigger done-events
                return

            self.__save_data.save(is_auto_save=True)    # persist the progress

            def show_end_message(end_message: Optional[Message]):
                def _show_proceed_summary():
                    CommonQuestions.proceed_summary(self.__cur_map.name, new_level_data.score, new_level_data.duration,
                                                    new_level_data.total_score, self.__proceed,
                                                    None if prev_level_data is None
                                                    else (prev_level_data.total_score, prev_level_data.duration))
                # either show the end message followed by the summary or just the summary
                if end_message is None: _show_proceed_summary()
                else: Popup.from_message_trigger(end_message, _show_proceed_summary)
            self.__cur_map.end(show_end_message)

        elif event_id.lower().startswith(MapConfig.unlock_prefix()):
            self.__save_data.unlock(event_id[len(MapConfig.unlock_prefix()):])
        else:
            if event_id in self.__temp_level_event_storage:  # todo: is score needed? a simple flag might be better
                event_score, event_done_score = self.__temp_level_event_storage[event_id]
                if event_score < event_done_score:
                    Config.check_reachability("MapManager.trigger_event() for existing event")
                    self.__temp_level_event_storage[event_id] = event_score + 1, event_done_score
            else:
                self.__temp_level_event_storage[event_id] = 1, 1

    def check_level_event(self, event: str) -> bool:
        """
        Used for popups to correctly display event-based messages.
        """
        if event in self.__temp_level_event_storage:
            score, done_score = self.__temp_level_event_storage[event]
            return score >= done_score
        return False

    def on_level_start(self):
        self.__temp_level_event_storage.clear()
        self.__level_timer = cur_datetime()

    def load_first_uncleared_map(self) -> None:
        seed = self.__rm.get_seed("MapManager.load_first_uncleared_map()")
        if Config.test_level(ignore_debugging=False):
            self.load_map(MapConfig.test_level(), None, seed)
        else:
            self.load_map(MapConfig.first_uncleared(), None, seed)

    def reload(self):
        if self.is_in_level:
            self.__load_level(self.__cur_map.internal_name, seed=None)     # set seed None to re-randomize it
        elif self.is_in_expedition:
            exp_map: ExpeditionMap = self.__cur_map
            # use same difficulty and map_seed but re-randomize puzzle_seed
            self.__load_expedition(exp_map.difficulty, exp_map.seed, None)
        else:
            Popup.error(f"Can only reload levels and expeditions, but the current map "
                        f"\"{self.__cur_map.internal_name}\" is neither. Please exit the map.", add_report_note=True)



from typing import List, Optional, Union

import numpy as np

from qrogue.util import Logger, QuantumSimulationConfig, GameplayConfig, Options, Config
from qrogue.util.quantum_functions import generate_ket, verify_circuit_matrix
from qrogue.util.util_functions import center_string, complex2string
from .state_vector import StateVector


class CircuitMatrix:
    @staticmethod
    def create_identity(num_of_qubits: int) -> "CircuitMatrix":
        matrix = []
        for i in range(2 ** num_of_qubits):
            row = [0] * 2 ** num_of_qubits
            row[i] = 1
            matrix.append(row)
        return CircuitMatrix(matrix, num_of_used_gates=0)

    @staticmethod
    def check_validity(matrix: List[List[complex]]) -> bool:
        col_sum: List[float] = [0] * len(matrix)
        for row in matrix:
            row_sum = 0
            for i, val in enumerate(row):
                res_val = abs(val) ** 2
                row_sum += res_val
                col_sum[i] += res_val
            if row_sum < 1 - QuantumSimulationConfig.TOLERANCE or 1 + QuantumSimulationConfig.TOLERANCE < row_sum:
                return False
        for val in col_sum:
            if val < 1 - QuantumSimulationConfig.TOLERANCE or 1 + QuantumSimulationConfig.TOLERANCE < val:
                return False
        return True

    @staticmethod
    def matrix_to_string(matrix: Union[List[List[complex]], np.ndarray], num_of_qubits: int,
                         space_per_value: Optional[int] = None) -> str:
        if isinstance(matrix, np.ndarray):
            assert len(matrix.shape) == 2 and matrix.shape[0] == matrix.shape[1], \
                f"matrix needs to be 2D and quadratic, but is: {matrix.shape}"
        if space_per_value is None:
            # use the maximum space by default, so we definitely have enough space
            space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_COMPLEX_NUMBER

        spacing = " "
        if GameplayConfig.get_option_value(Options.show_ket_notation, convert=True):
            padding = len(generate_ket(0, num_of_qubits)) + len(spacing)  # also add the space after the ket
            text = " " * padding  # we need to pad the rows' |qubits> prefix
            for i in range(2 ** num_of_qubits):
                # space_per_value + 1 due to the trailing space
                text += center_string(generate_ket(i, num_of_qubits), space_per_value)
                text += spacing
            text += "\n"
        else:
            text = "\n"
        for i, row in enumerate(matrix):
            if GameplayConfig.get_option_value(Options.show_ket_notation, convert=True):
                text += generate_ket(i, num_of_qubits)
                text += spacing
            for val in row:
                decimals = QuantumSimulationConfig.COMPLEX_DECIMALS if val.real != 0 and val.imag != 0 \
                    else QuantumSimulationConfig.DECIMALS
                text += center_string(complex2string(val, decimals), space_per_value)
                text += spacing
            text += "\n"
        return text

    def __init__(self, matrix: List[List[complex]], num_of_used_gates: int):
        self.__matrix = matrix
        self.__num_of_used_gates = num_of_used_gates

        if not verify_circuit_matrix(matrix, QuantumSimulationConfig.TOLERANCE):
            Config.check_reachability("CircuitMatrix.verfiy_circuit_matrix()")

    @property
    def size(self) -> int:
        return len(self.__matrix)

    @property
    def num_of_qubits(self) -> int:
        return int(np.log2(self.size))

    @property
    def num_of_used_gates(self) -> int:
        return self.__num_of_used_gates

    @property
    def is_classical(self) -> bool:
        mat_sum = 0
        for row in self.__matrix:
            for val in row:
                if val != 0 and val != 1:
                    return False
                mat_sum += val
        return mat_sum == self.size  # there needs to be one 1 per row

    @property
    def is_real(self) -> bool:
        for row in self.__matrix:
            for val in row:
                if val.imag != 0:
                    return False
        return True

    @property
    def is_imag(self) -> bool:
        for row in self.__matrix:
            for val in row:
                if val.real != 0:
                    return False
        return True

    @property
    def is_complex(self) -> bool:
        # if it's neither pure real nor pure imaginary it has to be mixed and therefore complex
        return not self.is_real and not self.is_imag

    def multiply(self, stv: StateVector) -> Optional[StateVector]:  # todo improve performance?
        if self.num_of_qubits == stv.num_of_qubits:
            values = []
            for row in self.__matrix:
                val = 0
                for i, entry in enumerate(row):
                    val += entry * stv.at(i)
                values.append(val)
            return StateVector(values, self.num_of_used_gates +
                               (0 if stv.num_of_used_gates is None else stv.num_of_used_gates))
        else:
            Logger.instance().error(f"@multiply: CircuitMatrix (={self.num_of_qubits}) and Stv (={stv.num_of_qubits}) "
                                    f"don't have the same number of qubits!", show=Config.debugging(), from_pycui=False)
            return None

    def to_string(self, space_per_value: Optional[int] = None) -> str:
        if space_per_value is None:
            if self.is_complex:
                space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_COMPLEX_NUMBER
            elif self.is_imag:
                space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_NUMBER + 1
            else:
                space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_NUMBER

        return CircuitMatrix.matrix_to_string(self.__matrix, self.num_of_qubits, space_per_value)

    def __eq__(self, other):
        if not isinstance(other, CircuitMatrix): return False
        if other.size != self.size: return False

        for y in range(self.size):
            for x in range(self.size):
                if other.__matrix[y][x] != self.__matrix[y][x]:
                    return False
        return True

    def __str__(self) -> str:
        text = "CircuitMatrix("
        for row in self.__matrix:
            for val in row:
                text += f"{np.round(val, QuantumSimulationConfig.DECIMALS)}, "
            text += "\n"
        text = text[:-2] + ")"
        return text



# exporting
import itertools
from typing import Tuple, Optional, List

from qrogue.game.logic.base import CircuitMatrix, StateVector, QuantumCircuit, UnitarySimulator
from qrogue.game.logic.collectibles import Instruction
from qrogue.util import QuantumSimulationConfig
from .boss import Boss
from .challenge import Challenge
from .enemy import Enemy
from .fusion_target import FusionTarget
from .riddle import Riddle
from .target import Target


def is_puzzle_solvable(input_stv: StateVector, target_stv: StateVector, gate_list: List[Instruction]) \
        -> Tuple[bool, Optional[QuantumCircuit]]:
    # todo: sometimes the order of the gates matters and claims that a solvable puzzle is not solvable
    #  Example:
    #  input_stv = StateVector([0, 0, 0, 0, complex(0, -1), 0, 0, 0])
    #  target_stv = Instruction.compute_stv([gates.HGate().setup([1])], 3)
    #  gate_list = [gates.XGate(), gates.CXGate(), gates.HGate(), gates.SGate()]     # this order returns True
    #  gate_list_ = [gates.HGate(), gates.SGate(), gates.XGate(), gates.CXGate()]      # but this order returns False

    assert input_stv.num_of_qubits == target_stv.num_of_qubits, "Different number of qubits!"

    simulator = UnitarySimulator()
    circuit = QuantumCircuit.from_bit_num(input_stv.num_of_qubits, input_stv.num_of_qubits)
    return __is_reachable_rec(simulator, circuit, input_stv, target_stv, gate_list)


def __is_reachable_rec(simulator: UnitarySimulator, circuit: QuantumCircuit, input_stv: StateVector,
                       target_stv: StateVector, gate_list: List[Instruction]) \
        -> Tuple[bool, Optional[QuantumCircuit]]:
    cur_gate = gate_list.pop()
    qubits = list(itertools.permutations(range(input_stv.num_of_qubits), cur_gate.num_of_qubits))

    while len(qubits) > 0:
        cur_circuit = circuit.copy()
        qargs = qubits.pop()
        gate = cur_gate.copy().setup(qargs)
        gate.append_to(cur_circuit)

        amplitudes = simulator.execute(cur_circuit, decimals=QuantumSimulationConfig.DECIMALS)
        circuit_matrix = CircuitMatrix(amplitudes, 0)  # num of used gates doesn't matter when checking reachability
        cur_stv = circuit_matrix.multiply(input_stv)
        diff = cur_stv.get_diff(target_stv)

        if diff.is_zero:
            return True, cur_circuit

        # go one recursion step deeper and try the next gate
        if len(gate_list) > 0:
            ret, winning_circuit = __is_reachable_rec(simulator, cur_circuit, input_stv, target_stv,
                                                      gate_list.copy())
            if ret: return True, winning_circuit
        # else we try to place cur_gate on different qubits and continue

    return False, None

# importing
# +util
# +collectibles
# +StateVector



from typing import Optional, Callable, Tuple


class Message:
    __err_count = 0

    @staticmethod
    def error(text: str) -> "Message":
        Message.__err_count += 1
        return Message(f"Error_{Message.__err_count}", "Error", text, False, None, None, None)

    @staticmethod
    def create_with_title(m_id: str, title: str, text: str, priority: bool, position: Optional[int] = None) \
            -> "Message":
        return Message(m_id, title, text, priority, position, None, None)

    @staticmethod
    def create_with_alternative(m_id: str, title: str, text: str, priority: bool, position: int, event: str,
                                alternative: "Message") -> "Message":
        message = Message(m_id, title, text, priority, position, event, alternative.id)
        message.resolve_message_ref(alternative)
        return message

    @staticmethod
    def create_with_exception(m_id: str, title: str, text: str, priority: bool, event: str):
        return Message(m_id, title, text, priority, None, event, "none")

    @staticmethod
    def create_from_message(message: "Message") -> "Message":
        msg = Message(message.__m_id, message.__title, message.__text, message.__priority, message.__position,
                      message.__event, message.__alt_ref)
        if message.__alt_message:
            msg.resolve_message_ref(message.__alt_message)
        return msg

    def __init__(self, m_id: str, title: str, text: str, priority: bool, position: Optional[int], event: Optional[str],
                 alternative: Optional[str]):
        assert position is None or isinstance(position, int), "Invalid position type!"

        self.__m_id = m_id
        self.__title = title
        self.__text = text
        self.__priority = priority
        self.__position = position
        self.__alt_ref = alternative
        self.__event = event
        self.__alt_message = None

        if self.__alt_ref is self.id:
            raise ValueError("Messages must not reference themselves as alternatives!")

    @property
    def id(self) -> str:
        return self.__m_id

    @property
    def priority(self) -> bool:
        return self.__priority

    @property
    def position(self) -> Optional[int]:
        return self.__position

    @property
    def alt_message_ref(self) -> Optional[str]:
        return self.__alt_ref

    def resolve_message_ref(self, message: "Message") -> bool:
        if self.alt_message_ref == message.__m_id:
            # check for cycles
            alt = message
            length = 1
            while alt:
                if alt is self:
                    raise ValueError(f"Found cycle of length {length} in alternative messages!")
                alt = alt.__alt_message
                length += 1
            self.__alt_message = message
            return True
        return False

    def get(self, check_achievement: Callable[[str], bool]) -> Optional[Tuple[str, str]]:
        if self.__event is not None and check_achievement(self.__event):
            if self.alt_message_ref == "none":
                return None
            if self.__alt_message:
                return self.__alt_message.get(check_achievement)
        else:
            return self.__title, self.__text



from abc import ABC, abstractmethod
from enum import Enum
from typing import List

from qrogue.game.logic.actors import Controllable
from qrogue.game.world.navigation import Direction
from qrogue.util import CheatConfig


class TileCode(Enum):
    Invalid = (-1, "§")         # when an error occurs, e.g. a tile at a non-existing position should be retrieved
    Debug = (-2, "€")           # displays a digit for debugging
    Void = (7, "_")             # tile outside the playable area
    Floor = (0, " ")            # simple floor tile without special meaning
    HallwayEntrance = (5, " ")  # depending on the hallway it refers to is either a Floor or Wall
    FogOfWar = (3, "~")         # tile of a place we cannot see yet

    Message = (6, ".")          # tile for displaying a popup message
    Goal = (61, "G")            # tile for the goal, i.e. end of a level
    Trigger = (9, "T")          # tile that calls a function on walk, i.e. event tile
    Teleport = (91, "t")        # special trigger for teleporting between maps
    Decoration = (11, "d")      # simply displays a specified character

    Wall = (1, "#")
    Obstacle = (2, "o")
    Door = (4, "|")

    Controllable = (20, "C")
    # Npc = (25, "N")
    Enemy = (30, "E")
    Boss = (40, "B")

    Collectible = (50, "c")
    Riddler = (51, "?")
    # ShopKeeper = (52, "$")
    Energy = (53, "e")
    Challenger = (54, "!")

    # SpaceshipBlock = (70, "/")
    # SpaceshipWalk = (71, ".")
    # SpaceshipTrigger = (72, "T")
    # OuterSpace = (73, "*")

    CollectibleKey = (501, "k")
    # CollectibleCoin = (502, "€")
    CollectibleEnergy = (503, "e")
    CollectibleScore = (504, "s")
    CollectibleGate = (520, "g")
    CollectibleQubit = (560, "q")

    def __init__(self, id: int, representation: str):
        self.__id = id
        self.__representation = representation

    @property
    def id(self) -> int:
        return self.__id

    @property
    def representation(self) -> str:
        return self.__representation

    def __str__(self):
        return self.representation

    @staticmethod
    def special_values() -> List["TileCode"]:
        return [
            TileCode.Invalid, TileCode.Debug, TileCode.Void, TileCode.FogOfWar,
            TileCode.Message, TileCode.Trigger, TileCode.Teleport, TileCode.Decoration,
        ]

    @staticmethod
    def collectible_subtypes() -> List["TileCode"]:
        return [
            TileCode.Collectible,
            TileCode.CollectibleKey, TileCode.CollectibleEnergy,
            TileCode.CollectibleGate, TileCode.CollectibleQubit,
            TileCode.CollectibleScore,
        ]


class Tile(ABC):
    @staticmethod
    def _invisible_tile():
        return " "

    def __init__(self, code: TileCode):
        self.__code = code

    @property
    def data(self) -> None:
        """
        E.g. amount for Energy, logical Collectible for Collectibles, eid for Enemies or None for tiles without any
        dynamic values (Walls, Floor, ...)

        :return: a representation of the specific tile's data
        """
        return None

    @property
    def code(self) -> TileCode:
        return self.__code

    @property
    def _invisible(self):
        return Tile._invisible_tile()

    @abstractmethod
    def get_img(self):
        pass

    @abstractmethod
    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        pass

    @abstractmethod
    def copy(self) -> "Tile":
        pass

    def __str__(self):
        return self.get_img()


class Invalid(Tile):
    def __init__(self):
        super().__init__(TileCode.Invalid)

    @property
    def data(self) -> None:
        return None

    def get_img(self):
        return TileCode.Invalid.representation

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False

    def copy(self) -> "Tile":
        return Invalid()


class Debug(Tile):
    def __init__(self, num: int):
        super(Debug, self).__init__(TileCode.Debug)
        self.__num = str(num)[0]

    @property
    def data(self) -> int:
        return int(self.__num)

    def get_img(self) -> str:
        return self.__num

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False

    def copy(self) -> "Tile":
        return Debug(int(self.__num))


class Void(Tile):
    def __init__(self):
        super().__init__(TileCode.Floor)

    @property
    def data(self) -> None:
        return None

    def get_img(self):
        return self._invisible

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False

    def copy(self) -> "Tile":
        return Void()


class Floor(Tile):
    @staticmethod
    def img():
        return Tile._invisible_tile()

    def __init__(self):
        super().__init__(TileCode.Floor)

    @property
    def data(self) -> None:
        return None

    def get_img(self):
        return Floor.img()

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return True

    def copy(self) -> "Tile":
        return Floor()


class Wall(Tile):
    @staticmethod
    def img():
        return TileCode.Wall.representation

    def __init__(self):
        super().__init__(TileCode.Wall)

    @property
    def data(self) -> None:
        return None

    def get_img(self):
        return Wall.img()

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False

    def copy(self) -> "Tile":
        return Wall()


class Obstacle(Tile):
    def __init__(self):
        super().__init__(TileCode.Obstacle)

    @property
    def data(self) -> None:
        return None

    def get_img(self):
        return TileCode.Obstacle.representation

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False or CheatConfig.ignore_obstacles()

    def copy(self) -> "Tile":
        return Obstacle()


class FogOfWar(Tile):
    def __init__(self):
        super().__init__(TileCode.Obstacle)

    @property
    def data(self) -> None:
        return None

    def get_img(self):
        return TileCode.FogOfWar.representation

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return True

    def copy(self) -> "Tile":
        return FogOfWar()


class Decoration(Tile):
    def __init__(self, decoration: str, blocking: bool = False):
        super(Decoration, self).__init__(TileCode.Decoration)
        self.__decoration = decoration
        self.__blocking = blocking

    @property
    def data(self) -> bool:
        return self.__blocking

    def get_img(self):
        return self.__decoration

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return not self.__blocking

    def copy(self) -> "Tile":
        return Decoration(self.__decoration, self.__blocking)


class ControllableTile(Tile):
    def __init__(self, controllable: Controllable):
        super().__init__(TileCode.Controllable)
        self.__controllable = controllable

    @property
    def data(self) -> str:
        return self.__controllable.name

    def get_img(self):
        return self.__controllable.get_img()

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return True  # todo check

    @property
    def controllable(self) -> Controllable:
        return self.__controllable

    def copy(self) -> "Tile":
        return ControllableTile(self.__controllable)



# exporting
from .my_widgets import MyBaseWidget, Widget
from .renderable import Renderable
from .widget_sets import ExploreWidgetSet, NavigationWidgetSet
from .widget_sets import MenuWidgetSet, TransitionWidgetSet, PauseMenuWidgetSet, WorkbenchWidgetSet, \
    ScreenCheckWidgetSet, LevelSelectWidgetSet
from .widget_sets import MyWidgetSet
from .widget_sets import ReachTargetWidgetSet, TrainingsWidgetSet, FightWidgetSet, BossFightWidgetSet, \
    RiddleWidgetSet, ChallengeWidgetSet, FusionCircuitWidgetSet

# importing
# +game.logic.actors
# +game.logic.actors.puzzles
# +game.logic.base
# +game.logic.collectibles
# +game.world.map
# +game.world.navigation
# +graphics.popups
# +graphics.rendering
# +util



from typing import List, Tuple, Optional

from qrogue.game.logic.base import StateVector, CircuitMatrix
from qrogue.game.logic.collectibles import Collectible, Instruction
from qrogue.util import CheatConfig, Logger, Config
from .target import Target


class Challenge(Target):
    def __init__(self, id_: int, target: StateVector, min_gates: int, max_gates: Optional[int] = None,
                 reward: Optional[Collectible] = None, allowed_gates: List[Instruction] = None,
                 input_: Optional[StateVector] = None):
        if max_gates < min_gates:
            Logger.instance().error(f"Trying to create a challenge with less max_gates={max_gates} then "
                                    f"min_gates={min_gates}! Using min_gates for both values instead.", show=False,
                                    from_pycui=False)
        if max_gates is None:
            max_gates = min_gates
        # allow target and input to be equal since other constraints can still make it challenging
        super().__init__(id_, target, reward, input_, allow_target_input_equality=True)
        self.__min_gates = min_gates
        self.__max_gates = max_gates
        self.__allowed_gates = allowed_gates  # not yet usable!     # todo: make "mandatory_gates" instead?

    @property
    def flee_energy(self) -> int:
        return 10

    @property
    def min_gates(self) -> int:
        return self.__min_gates

    @property
    def max_gates(self) -> int:
        return self.__max_gates

    def is_reached(self, state_vector: StateVector, circ_matrix: CircuitMatrix) -> Tuple[bool, Optional[Collectible]]:
        # first check if the additional constraints are met
        if circ_matrix.num_of_used_gates is None:
            Logger.instance().error(f"Challenge.is_reached() got state_vector with no number of used gates specified!",
                                    show=Config.debugging(), from_pycui=False)
            # we cannot check the num_of_gates constraint
            return super(Challenge, self).is_reached(state_vector, circ_matrix)

        if self.__min_gates <= circ_matrix.num_of_used_gates <= self.__max_gates or CheatConfig.in_god_mode():
            return super(Challenge, self).is_reached(state_vector, circ_matrix)
        return False, None



from typing import Dict, Optional, Tuple

from qrogue.game.logic import Message
from qrogue.game.world.map import Hallway
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.game.world.tiles import Door
from qrogue.util import MapConfig, Logger, Config

DEFAULT_HALLWAY_STR = "=="
TEMPLATE_PREFIX = "_"
EMPTY_ROOM_CODE = "_a"
EMPTY_HALLWAY_CODE = "_0"

# how the tiles look in the editor is not necessarily the same as they look in the game
OBSTACLE_TILE = "o"
PLACE_HOLDER_TILE = "_"
COLLECTIBLE_TILE = "c"
TRIGGER_TILE = "t"
TELEPORT_TILE = "t"
MESSAGE_TILE = "m"
ENERGY_TILE = "e"
BOSS_TILE = "b"
RIDDLER_TILE = "r"
CHALLENGER_TILE = "!"
FLOOR_TILE = " "


def warning(text: str, location: str):
    Logger.instance().warn(f"@{location}: {text}", from_pycui=False)


def error(text: str):
    Logger.instance().error(text, show=False, from_pycui=False)
    raise SyntaxError(text)


def check_for_overspecified_columns(x: int, symbol_type, ref_type):
    return x == MapConfig.map_width() and symbol_type != ref_type or x > MapConfig.map_width()


def normalize_reference(reference: str) -> str:
    if reference[0] == '*':
        return reference[1:].lower()
    else:
        return reference.lower()


def direction_from_string(dir_str: str) -> Direction:
    direction = None
    if dir_str == "North":
        direction = Direction.North
    elif dir_str == "East":
        direction = Direction.East
    elif dir_str == "South":
        direction = Direction.South
    elif dir_str == "West":
        direction = Direction.West
    return direction


def get_hallways(hallway_dict: Dict[Coordinate, Dict[Direction, Hallway]],
                 door_dict: Dict[Coordinate, Dict[Coordinate, Door]], pos: Coordinate) \
        -> Optional[Dict[Direction, Hallway]]:
    if pos in door_dict:
        doors = door_dict[pos]
        if doors:
            room_hallways = {
                Direction.North: None, Direction.East: None, Direction.South: None, Direction.West: None,
            }
            for neighbor in doors:
                direction = Direction.from_coordinates(pos, neighbor)
                opposite = direction.opposite()
                # get hallway from neighbor if it exists, otherwise create it
                if neighbor in hallway_dict and opposite in hallway_dict[neighbor]:
                    hallway = hallway_dict[neighbor][opposite]
                else:
                    door = doors[neighbor]
                    # always make a copy, so we don't run into problems if we use a simple hallway multiple times
                    if door.is_one_way:
                        if door.direction not in [direction, direction.opposite()]:
                            warning("Found one way door with invalid direction! Horizontal hallways can only "
                                    "have East or West doors and vertical ones only North and South doors! "
                                    "Removing the one way aspect and setting the direction to a valid one.")
                            door = door.copy_and_adapt(direction, reset_one_way=True)
                        else:
                            door = door.copy()
                    else:
                        door = door.copy_and_adapt(direction)
                    hallway = Hallway(door)
                    if neighbor in hallway_dict:
                        hallway_dict[neighbor][opposite] = hallway
                    else:
                        hallway_dict[neighbor] = {opposite: hallway}

                # store the hallway so the neighbors can find it if necessary
                if pos not in hallway_dict:
                    hallway_dict[pos] = {}
                hallway_dict[pos][direction] = hallway
                room_hallways[direction] = hallway
            return room_hallways
    return None


def text_to_str(ctx_text, index: int = None) -> str:
    if index is None:
        text = ctx_text.TEXT()
    else:
        text = ctx_text.TEXT(index)
    return text.getText()[1:-1]


def parse_integer(ctx) -> Optional[int]:
    if ctx.DIGIT():
        return int(ctx.DIGIT().getText())
    elif ctx.HALLWAY_ID():
        return int(ctx.HALLWAY_ID().getText())
    elif ctx.INTEGER():
        return int(ctx.INTEGER().getText())
    else:
        return None


def parse_complex(ctx) -> complex:
    if ctx.SIGN(0):
        first_sign = ctx.SIGN(0).symbol.text
    else:
        first_sign = "+"

    integer_ = ctx.integer()
    float_ = ctx.FLOAT()
    imag_ = ctx.IMAG_NUMBER()

    complex_number = first_sign
    if integer_ or float_:
        if integer_:
            num = str(parse_integer(integer_))
        else:
            num = float_.getText()
        complex_number += num

        if ctx.SIGN(1):
            complex_number += ctx.SIGN(1).symbol.text + str(imag_)
    else:
        complex_number += str(imag_)

    return complex(complex_number)


def parse_message(ctx, default_speaker: str = Config.system_name()) -> Message:
    m_id = normalize_reference(ctx.REFERENCE(0).getText())

    title, priority, position, msg = parse_message_body(ctx.message_body(), default_speaker)

    if ctx.MSG_EVENT():
        event = normalize_reference(ctx.REFERENCE(1).getText())
        msg_ref = normalize_reference(ctx.REFERENCE(2).getText())
        return Message(m_id, title, msg, priority, position, event, msg_ref)
    else:
        return Message.create_with_title(m_id, title, msg, priority, position)


def parse_speaker(ctx, text_index: Optional[int] = 0) -> str:
    title = text_to_str(ctx, text_index)
    if title.isdigit():
        title = Config.get_name(int(title[0]))
    return title


def parse_message_body(ctx, default_speaker: str) -> Tuple[str, bool, int, str]:
    if ctx.MSG_SPEAKER():
        title = parse_speaker(ctx, 0)
        start = 1
    else:
        title = default_speaker
        start = 0
    priority = ctx.MSG_PRIORITY() is not None
    if ctx.DIGIT():
        position = int(ctx.DIGIT().getText())
    else:
        position = None
    msg = ""
    for i in range(start, len(ctx.TEXT())):
        msg += text_to_str(ctx, i) + "\n"

    return title, priority, position, msg



from typing import Any, Dict, Optional, List, Iterator

from qrogue.game.world.dungeon_generator.wave_function_collapse import WFCLearnMatrix
from qrogue.game.world.navigation import Coordinate, Direction


class WFCLearner:
    def __init__(self, pos_weights: Dict[Coordinate, Dict[Any, int]],
                 type_weights: Dict[Optional[Any], Dict[Any, int]]):
        self.__pos_weights = pos_weights
        self.__type_weights = type_weights

        self.__width = -1
        self.__height = -1
        for coor in self.__pos_weights:
            if coor.x > self.__width:
                self.__width = coor.x
            if coor.y > self.__height:
                self.__height = coor.y

        # increment width and height since coordinates start at 0
        self.__width += 1
        self.__height += 1
        # if pos_weights was empty, width and height are now 0

    @property
    def width(self) -> int:
        return self.__width

    @property
    def height(self) -> int:
        return self.__height

    @property
    def positions(self) -> Iterator[Coordinate]:
        return iter(self.__pos_weights.keys())

    @property
    def types(self) -> Iterator[Optional[Any]]:
        return iter(self.__type_weights.keys())

    def pos_weights(self, pos: Coordinate) -> Dict[Any, int]:
        if pos in self.__pos_weights:
            return self.__pos_weights[pos].copy()
        return {}

    def type_weights(self, type_: Optional[Any]):
        if type_ in self.__type_weights:
            return self.__type_weights[type_].copy()
        return {}

    def learn(self, matrix: WFCLearnMatrix):
        if self.__width > 0:
            if matrix.width != self.__width:
                from qrogue.util import Config
                Config.check_reachability("WFCLearner.learn() inequal width")
                # todo: Log error?
        else:
            self.__width = matrix.width

        if self.__height > 0:
            if matrix.height != self.__height:
                from qrogue.util import Config
                Config.check_reachability("WFCLearner.learn() inequal height")
                # todo: Log error?
        else:
            self.__height = matrix.height

        for x in range(matrix.width):
            for y in range(matrix.height):
                value = matrix.at(x, y)

                # update positional weights
                c = Coordinate(x, y)
                if c not in self.__pos_weights:
                    self.__pos_weights[c] = {}
                if value not in self.__pos_weights[c]:
                    self.__pos_weights[c][value] = 0
                self.__pos_weights[c][value] += 1

                # update type weights
                if value not in self.__type_weights:
                    self.__type_weights[value] = {}
                neighbors = [matrix.at(x + direction.x, y + direction.y)
                             for direction in Direction.values()]
                for neigh in neighbors:
                    if neigh not in self.__type_weights[value]:
                        self.__type_weights[value][neigh] = 0
                    self.__type_weights[value][neigh] += 1

    def remove_unwanted_values(self) -> None:
        """
        Can be overwritten to define values that we don't want to learn from.

        :return: None
        """
        pass

    def _remove_unwanted_values(self, values: List[Any]):
        """
        Can be used by subclasses to easily remove the listed values from the learned weights

        :param values:
        :return:
        """
        for pos in self.__pos_weights:
            for tc in self.__pos_weights[pos]:
                if tc in values:
                    # by setting count to 0 we "delete" it without messing up the loop
                    self.__pos_weights[pos][tc] = 0

        for tc in values:
            if tc in self.__type_weights:
                self.__type_weights.pop(tc)



# exporting
import qrogue.game.world.dungeon_generator
import qrogue.game.world.map
import qrogue.game.world.navigation
import qrogue.game.world.tiles

# importing
# +graphics.popups
# +logic
# +logic.actors
# +logic.actors.controllables
# +logic.actors.puzzles
# +logic.base
# +logic.collectibles
# +util



from typing import Iterator, List, Optional

import numpy as np

from qrogue.util import Logger, QuantumSimulationConfig, GameplayConfig, Options, CheatConfig
from qrogue.util.config import ColorCode, ColorConfig
from qrogue.util.quantum_functions import generate_ket, verify_stv_amplitudes
from qrogue.util.util_functions import center_string, align_string, complex2string


def _wrap_in_ket_notation(number: complex, qubit: int, num_of_qubits: int,
                          decimals: Optional[int] = None, space_per_value: Optional[int] = None, coloring: bool = False,
                          correct_amplitude: bool = False, show_percentage: bool = False, skip_ket: bool = False):
    """

    :param number:
    :param qubit:
    :param num_of_qubits:
    :param decimals:
    :param space_per_value:
    :param coloring:
    :param correct_amplitude:
    :param show_percentage:
    :param skip_ket:
    :return:
    """
    is_complex = number.real != 0 and number.imag != 0
    if decimals is None:
        if is_complex:
            decimals = QuantumSimulationConfig.COMPLEX_DECIMALS
        else:
            decimals = QuantumSimulationConfig.DECIMALS

    if space_per_value is None:
        if is_complex:
            space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_COMPLEX_NUMBER
        elif number.imag != 0:
            space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_NUMBER + 1  # add the extra "j"
        else:
            space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_NUMBER

    value = f"{center_string(complex2string(number, decimals), space_per_value)}"
    if coloring:
        if correct_amplitude:
            value = ColorConfig.colorize(ColorCode.PUZZLE_CORRECT_AMPLITUDE, value)
        else:
            value = ColorConfig.colorize(ColorCode.PUZZLE_WRONG_AMPLITUDE, value)

    if show_percentage:
        space = QuantumSimulationConfig.MAX_PERCENTAGE_SPACE
        percentage = StateVector.complex_to_amplitude_percentage_string(number, space)
        space += 1  # +1 because of the additional "%" we have to align
        if percentage == "0%":
            value += f"  {align_string('', space, left=False)}"  # show whitespace instead of redundant 0% amplitudes
        else:
            value += f" ~{align_string(percentage, space, left=False)}"

    if GameplayConfig.get_option_value(Options.show_ket_notation, convert=True) and not skip_ket:
        return f"{generate_ket(qubit, num_of_qubits)}  {value}"
    else:
        return value


class StateVector:
    @staticmethod
    def check_amplitudes(amplitudes: List[complex]):
        return verify_stv_amplitudes(amplitudes, QuantumSimulationConfig.TOLERANCE)

    @staticmethod
    def complex_to_amplitude_percentage_string(val: complex,
                                               space: int = QuantumSimulationConfig.MAX_PERCENTAGE_SPACE) -> str:
        amp = np.round(abs(val ** 2), QuantumSimulationConfig.DECIMALS)
        text = str(amp * 100)
        if text[-2:] == ".0":
            text = text[:-2]  # remove the redundant ".0"
        if len(text) > space:
            text = text[:space]  # clamp to specified space
        return text + "%"

    @staticmethod
    def create_zero_state_vector(num_of_qubits: int) -> "StateVector":
        amplitudes = [1] + [0] * (2 ** num_of_qubits - 1)
        return StateVector(amplitudes, num_of_used_gates=0)

    @staticmethod
    def create_basis_states(num_of_qubits: int) -> List["StateVector"]:
        states = []
        for i in range(2 ** num_of_qubits):
            amplitudes = [0] * 2 ** num_of_qubits
            amplitudes[i] = 1
            states.append(StateVector(amplitudes, num_of_used_gates=0))
        return states

    def __init__(self, amplitudes: List[complex], num_of_used_gates: Optional[int] = None):
        self.__amplitudes = amplitudes
        self.__num_of_used_gates = num_of_used_gates

    @property
    def size(self) -> int:
        return len(self.__amplitudes)

    @property
    def num_of_qubits(self) -> int:
        return int(np.log2(self.size))

    @property
    def is_zero(self) -> bool:
        for val in self.__amplitudes:
            if abs(val) > QuantumSimulationConfig.TOLERANCE:
                return False
        return True

    @property
    def is_classical(self) -> bool:
        for val in self.__amplitudes:
            # if there is an imaginary part or the real part is anything except 0 or 1, the StateVector is not classical
            if val.imag != 0 or val.real not in [0, 1]:
                return False
        # if the sum is 1, we have exactly one 1 and everything else is 0 -> we have a classical state
        return sum(self.__amplitudes) == 1

    @property
    def is_real(self) -> bool:
        for val in self.__amplitudes:
            if val.imag != 0:
                return False
        return True

    @property
    def is_imag(self) -> bool:
        for val in self.__amplitudes:
            if val.real != 0:
                return False
        return True

    @property
    def is_complex(self) -> bool:
        # if it's neither pure real nor pure imaginary it has to be mixed and therefore complex
        return not self.is_real and not self.is_imag

    @property
    def num_of_used_gates(self) -> int:
        return self.__num_of_used_gates

    def at(self, index: int) -> complex:
        if 0 <= index < self.size:
            return self.__amplitudes[index]

    def to_value(self) -> List[float]:
        """
        Returns a list of amplitudes corresponding to this StateVector.

        :return: List of amplitudes
        """
        return [np.round(val.real ** 2 + val.imag ** 2, decimals=QuantumSimulationConfig.DECIMALS)
                for val in self.__amplitudes]

    def is_equal_to(self, other: "StateVector", tolerance: float = QuantumSimulationConfig.TOLERANCE,
                    ignore_god_mode: bool = False) -> bool:
        """

        :param other:
        :param tolerance:
        :param ignore_god_mode: whether we also want to check for equality when in god mode
        :return:
        """
        if not ignore_god_mode and CheatConfig.in_god_mode():
            return True

        # other needs at least as many entries as self_value, more are allowed
        #  (so the robot can have more qubits than the enemy)
        if self.size > other.size:
            return False
        """
        self_value = self.__amplitudes  #self.to_value()
        other_value = other.__amplitudes    #other.to_value()
        for i in range(len(self_value)):
            p_min = self_value[i] - tolerance/2         # todo maybe tolerance doesn't work for imaginary numbers?
            p = other_value[i]
            p_max = self_value[i] + tolerance/2
            if not (p_min <= p <= p_max):
                return False
        """
        diff = [self.__amplitudes[i] - other.__amplitudes[i] for i in range(self.size)]
        for val in diff:
            if abs(val) > tolerance:
                return False
        return True

    def get_diff(self, other: "StateVector") -> "StateVector":
        """
        Calculates the difference of this and other (i.e., this - other), with normal vector subtraction rules.
        To be more specific, the difference of every amplitude entry of this and other is calculated and then used to
        create a new StateVector.
        In the special case of this being smaller in size than other, this is first extended with 0s before the
        difference is calculated the usual way. This ensures that the resulting StateVector has always the same size as
        other.
        Should other be smaller in size than this, an erroneous StateVector with all -1 amplitudes is returned instead

        :param other: the StateVector we want to know the difference of
        :return: a new StateVector corresponding to this - other, with size equal to other
        """
        Logger.instance().assertion(self.size <= other.size, "Cannot calculate the difference between StateVectors of "
                                                             f"different sizes! self = {self}, other = {other}",
                                    show_popup=True)

        if self.size == other.size:
            diff = [self.__amplitudes[i] - other.__amplitudes[i] for i in range(self.size)]
            return StateVector(diff)

        elif self.size < other.size:
            Logger.instance().info("Requested difference between StateVectors of different sizes! "
                                   f"self = {self}, other = {other}; padding self with the needed number of 0s",
                                   from_pycui=False)
            diff: List[complex] = [0] * other.size
            for i in range(self.size):
                diff[i] = self.__amplitudes[i] - other.__amplitudes[i]
            return StateVector(diff)

        return StateVector([-1] * other.size)

    def wrap_in_qubit_conf(self, index: int, space_per_value: Optional[int] = None, coloring: bool = False,
                           correct_amplitude: bool = False, show_percentage: bool = False, skip_ket: bool = False) \
            -> str:
        # don't use default decimals since we don't want it to be dependent on individual entries but rather decide
        # based on whether the vector itself is complex or not
        decimals = QuantumSimulationConfig.COMPLEX_DECIMALS if self.is_complex else QuantumSimulationConfig.DECIMALS
        if space_per_value is None:
            if self.is_complex:
                space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_COMPLEX_NUMBER
            elif self.is_imag:
                space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_NUMBER + 1  # add the "j"
            else:
                space_per_value = QuantumSimulationConfig.MAX_SPACE_PER_NUMBER

        return _wrap_in_ket_notation(self.at(index), index, self.num_of_qubits, decimals, space_per_value, coloring,
                                     correct_amplitude, show_percentage, skip_ket)

    def to_string(self, space_per_value: Optional[int] = None) -> str:
        text = ""
        for i in range(self.size):
            text += self.wrap_in_qubit_conf(i, space_per_value)
            text += "\n"
        return text

    def __hash__(self):
        mult = 11
        hash_sum = 0
        for amplitude in self.__amplitudes:
            hash_sum += hash(amplitude.imag) * mult
            mult = mult * 2 + 1
            hash_sum += hash(amplitude.real) * mult
            mult = mult * 2 + 1
        return hash_sum

    def __eq__(self, other) -> bool:  # TODO currently not even in use!
        if type(other) is type(self):
            return self.__amplitudes == other.__amplitudes
        elif isinstance(other, list):
            if len(other) <= 0 or len(other) >= len(self.__amplitudes):
                return False
            if isinstance(other[0], bool):
                for i in range(len(self.__amplitudes)):
                    if self.__amplitudes[i] == 1 and not other[i] or self.__amplitudes[i] == 0 and other[i]:
                        return False
                return True
            elif isinstance(other[0], float):
                for i in range(len(self.__amplitudes)):
                    if self.__amplitudes[i] != other[i]:
                        return False
                return True
        return False

    def __str__(self) -> str:
        text = "StateVector("
        for val in self.__amplitudes:
            text += f"{np.round(val, QuantumSimulationConfig.DECIMALS)}, "
        text = text[:-2] + ")"
        return text

    def __iter__(self) -> Iterator:
        return iter(self.__amplitudes)



from typing import Callable, Optional, List, Union

from py_cui.debug import PyCUILogger

from .config import Config, ErrorConfig, PathConfig, TestConfig
from .util_functions import cur_datetime


class Logger(PyCUILogger):
    __BUFFER_SIZE = 4096
    __instance = None

    @staticmethod
    def print_to_console(message: str) -> None:
        time_str = cur_datetime().strftime("%H-%M-%S")
        text = f"{{Qrogue}}{time_str}: {message}"
        print(text)

    @staticmethod
    def instance() -> "Logger":
        if Logger.__instance is None:
            raise Exception(ErrorConfig.singleton_no_init("Logger"))
        return Logger.__instance

    @staticmethod
    def reset():
        if TestConfig.is_active():
            if Logger.__instance is None: return
            Logger.__instance.flush()  # flush before we reset to not lose data
            Logger.__instance = None
        else:
            raise TestConfig.StateException(ErrorConfig.singleton_reset("Logger"))

    @staticmethod
    def _set_instance(logger: "Logger"):
        # needed to insert a special TestLogger during testing
        Logger.__instance = logger

    def __init__(self, commit: Optional[Callable[[str], None]] = None):
        super().__init__("Qrogue-Logger")
        if Logger.__instance is not None:
            Logger.__instance.throw(Exception(ErrorConfig.singleton("Logger")))
        else:
            if commit is None:
                save_file = PathConfig.new_log_file()

                def commit_(text: str):
                    PathConfig.write(save_file, text, may_exist=True, append=True)

                self.__commit = commit_
            else:
                self.__commit = commit

            self.__text = ""
            self.__error_popup: Optional[Callable[[str], None]] = None
            self.__buffer: List[str] = []  # stores logged lines
            self.__error_counter = 0
            Logger.__instance = self

    @property
    def __buffer_size(self) -> int:
        """
        Returns: number of characters stored in buffer
        """
        return sum([len(line) for line in self.__buffer])

    @property
    def error_count(self) -> int:
        return self.__error_counter

    def set_popup(self, error_popup_function: Callable[[str], None]) -> None:
        self.__error_popup = error_popup_function

    def _write(self, text: str, from_pycui: Optional[bool]) -> None:
        """
        Args:
            text: the text to write to buffer
            from_pycui: True if text comes from PyCUI, False if it comes from QRogue, None if we do not know (e.g.,
                        some Exceptions)
        """
        self.__buffer.append(text)
        if self.__buffer_size >= Logger.__BUFFER_SIZE:
            self.flush()

    def info(self, message, from_pycui: bool = True, **kwargs) -> None:
        time_str = cur_datetime().strftime("%H-%M-%S")
        if from_pycui:
            self._write(f"{{PyCUI}}{time_str}: {message}", True)
        else:
            text = f"{{Qrogue}}{time_str}: {message}"
            self._write(text, False)
            if Config.debugging(): print(text)

    def debug(self, msg: str, from_pycui: bool = True, *args, **kwargs) -> None:
        if from_pycui and msg.startswith("Generated fragments"): return
        if Config.debugging():
            self.info(f"{{DEBUG}} |{msg}", from_pycui=from_pycui)

    def warn(self, msg: str, from_pycui: bool = True, *args, **kwargs) -> None:
        highlighting = "\n----------------------------------\n"
        self.info(f"{highlighting}WARNING |{msg}{highlighting}", from_pycui=from_pycui)

    def assertion(self, statement: bool, message, show_popup: bool = False):
        if Config.debugging():
            assert statement, str(message)
        else:
            if not statement:
                self.error(message, show_popup, from_pycui=False)

    def error(self, message, show: bool = True, from_pycui: bool = True, **kwargs) -> None:
        self.__error_counter += 1
        if show:
            self.__error_popup(str(message))
        highlighting = "\n----------------------------------\n"
        self.info(f"{highlighting}ERROR |{message}{highlighting}", from_pycui=from_pycui)
        Config.check_reachability("Logger.error()")

    def throw(self, error: Union[BaseException, str]) -> None:
        if isinstance(error, str):
            error = Exception(error)
        self.__error_counter += 1
        print(error)
        self._write(f"[ERROR] {error}", None)
        self.flush()
        Config.check_reachability("Logger.throw()")
        raise error

    def clear(self) -> None:
        self.__text = ""

    def flush(self) -> None:
        if self.__buffer_size > 0:
            text = ""
            for log in self.__buffer:
                text += log + "\n"
            self.__commit(text)
            self.__buffer.clear()
